# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
This modules provides StyleSheet, Style, Element and a some utility functions
that help write css properties.

:class:`StyleSheet` represents a list of rules and conditions (nested @-rules)
from a CSS file or string source. (The CSS format is parsed by the Css language
definition in :mod:`parce.lang.css` and transformed to a CSS structure by the
:class:`~parce.lang.css.CssTransform` transform class.)

:class:`Style` represents a resulting list of rules, sorted on specificity, so
that by selecting rules the properties that apply in a certain situation can be
determined and read.

:class:`Element` and :class:`AbstractElement` describe elements in a HTML or
XML document, and can be used to select matching rules in a stylesheet. Element
provides a list-based helper, and AbstractElement can be inherited from to wrap
any tree structure to use with stylesheet rule selectors.

This module is used by the :py:mod:`theme <parce.theme>` module to provide
syntax highlighting themes based on CSS files.

Workflow:

1. Instantiate a StyleSheet from a file or other source. If needed, combine
   multiple StyleSheets using the + operator.

2. Filter conditions out using ``filter_conditions()``, like media,
   supports or document.

3. Get a Style object through the ``style`` property of the StyleSheet.

4. Use a ``select`` method to select rules based on their selectors.

5. Use ``properties()`` to combine the properties of the selected rules to get
   a dictionary of the CSS properties that apply.

Example::

    >>> from parce.css import Element, StyleSheet
    >>> style = StyleSheet.from_file("parce/themes/default.css").style
    >>> e = Element(class_="comment", parent=Element(class_="parce"))
    >>> style.select_element(e).properties()
    {'color': [<Value text='dimgray', color=Color(r=105, g=105, b=105, a=1.0)>],
    'font-family': [<Value text='serif'>], 'font-style': [<Value text='italic'>]}

"""


import collections
import functools
import os
import re

from . import util
from .lang.css import Css, Atrule, Rule, Value
from .transform import transform_text


Condition = collections.namedtuple("Condition", "keyword node style")


def style_query(func):
    """Make a generator method return a new Style/StyleSheet/Atrules object."""
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        return type(self)(list(func(self, *args, **kwargs)))
    return wrapper


class StyleSheet:
    """Represents a list of style rules and conditions.

    Normal CSS rules are translated into a Rule tuple, nested rules
    such as @media, @document and @supports are translated into Condition
    tuples, and other @-rules are put in Atrule tuples.

    A Rule consists of ``selectors`` and ``properties``. The ``selectors``
    are the tokens in a rule before the {. The ``properties`` is a dict
    mapping css property names to the list of tokens representing their
    value.

    A Condition consists of ``keyword``, ``node`` and ``style``; the ``node``
    is Css.atrule context containing all text from the @ upto the opening {.
    The ``style`` is another StyleSheet object representing the nested
    style sheet.

    An Atrule tuple consists of ``keyword`` and ``node``, where the node is the
    Css.atrule context.

    You can combine stylesheets from different files or sources using the +
    operator.

    The ``style`` property returns the Style object representing all combined
    rules, and allowing further queries.  The ``at`` property returns an
    Atrules instance containing the atrules that do not belong to the nested
    at-rules.

    """
    filename = ""   #: our filename, if we were loaded from a file

    def __init__(self, rules=None, filename=""):
        """Initialize a StyleSheet, empty of with the supplied rules/conditions."""
        self.rules = rules or []
        self.filename = filename
        self._imported_filenames = []

    @classmethod
    def load_from_data(cls, data):
        """Return a Css.root tree from data, handling the encoding."""
        encoding, data = util.get_bom_encoding(data)
        if not encoding:
            m = re.match(rb'^@charset\s*"(.*?)"', data)
            encoding = m.group(1).decode('latin1') if m else "utf-8"
        try:
            text = data.decode(encoding)
        except (LookupError, UnicodeError):
            text = data.decode('utf-8', 'replace')
        return cls.load_from_text(text)

    @staticmethod
    def load_from_text(text):
        """Return a CSS structure from text."""
        return transform_text(Css.root, text)

    @classmethod
    def load_from_file(cls, filename):
        """Return a CSS structure from filename, handling the encoding."""
        return cls.load_from_data(open(filename, 'rb').read())

    @classmethod
    def from_file(cls, filename, path=None, allow_import=True):
        """Return a new StyleSheet adding Rules and Conditions from a local filename.

        The ``path`` argument is currently unused. If ``allow_import`` is
        False, the @import atrule is ignored.

        """
        css = cls.load_from_file(filename)
        return cls.from_css(css, filename, path, allow_import)

    @classmethod
    def from_text(cls, text, filename='', path=None, allow_import=True):
        """Return a new StyleSheet adding Rules and Conditions from a string.

        The ``filename`` argument is used to handle @import rules
        correctly. The ``path`` argument is currently unused. If
        ``allow_import`` is False, the @import atrule is ignored.

        """
        css = cls.load_from_text(text)
        return cls.from_css(css, filename, path, allow_import)

    @classmethod
    def from_data(cls, data, filename='', path=None, allow_import=True):
        """Return a new StyleSheet adding Rules and Conditions from a bytes string.

        The ``filename`` argument is used to handle @import rules
        correctly. The ``path`` argument is currently unused. If
        ``allow_import`` is False, the @import atrule is ignored.

        """
        css = cls.load_from_data(data)
        return cls.from_css(css, filename, path, allow_import)

    @classmethod
    def from_css(cls, css, filename='', path=None, allow_import=True):
        """Return a new StyleSheet adding Rules and Conditions from a CSS structure.

        The ``filename`` argument is used to handle @import rules
        correctly. The ``path`` argument is currently unused. If
        ``allow_import`` is False, the @import atrule is ignored.

        """
        filenames = {filename}

        def get_import_rules(values):
            """Yield rules from an @import at-rule.

            If the @import rule has a media query after the filename/url,
            one Condition is yielded.

            """
            for n, v in enumerate(values):
                if isinstance(v, Value) and v.text or v.url:
                    fname = v.text or v.url
                    fname = os.path.join(os.path.dirname(filename), fname)
                    # avoid circular @import references
                    if fname not in filenames:
                        filenames.add(fname)
                        icss = cls.load_from_file(fname)
                        values = values[n+1:]
                        if values:
                            # there is probably a media query after the filename
                            s = cls.from_css(icss, fname, path, allow_import)
                            yield Condition("import", values, s)
                        else:
                            self._imported_filenames.append(fname)
                            yield from get_rules(icss)
                    return

        def get_rules(css):
            """Get all CSS rules from the CSS structure, either as Rule or
            as Condition.

            The latter is used when rules can be selected or not depending
            on media, document, supports or import at-rules.

            """
            rules = []

            for rule in css:
                if isinstance(rule, Atrule):
                    # @-rule
                    if rule.keyword == "import":
                        if allow_import:
                            rules.extend(get_import_rules(rule.contents))
                        continue
                    elif isinstance(rule.block, list):
                        # nested @-rule
                        rule = Condition(rule.keyword, rule.contents, cls.from_css(rule.block))
                rules.append(rule)
            return rules

        return cls(get_rules(css), filename)

    def __add__(self, other):
        return type(self)(self.rules + other.rules)

    @style_query
    def filter_conditions(self, keyword, predicate):
        """Return a new StyleSheet object where conditions are filtered out.

        For Condition instances with the specified keyword, the predicate is
        called with the contents of the ``rule`` (the full Atrule) of
        each Condition, and if the return value doesn't evaluate to True, the
        Condition is removed from the resulting set. Conditions with other
        keywords are kept.

        Currently (CSS3), Conditions have the "media", "supports" or "document"
        keyword. @import rules that have a media query after the filename
        are also stored as a Condition.

        For example, this is a crude way to only get the @media rules for
        "screen"::

            filter_conditions("media", lambda rule: "screen" in rule.contents)

        Of course, a better parser for @media expressions could be written :-)

        """
        for r in self.rules:
            if isinstance(r, Condition) and r.keyword == keyword:
                if predicate(r):
                    yield r
            else:
                yield r

    def filenames(self):
        """Return a list of filenames the currently selected rules depend on.

        Our own filename will be the first in the list, and filenames of
        ``@import``-ed rules that are still selected are appended to the list.

        """
        def get_filenames(sheet):
            if sheet.filename:
                yield sheet.filename
            yield from sheet._imported_filenames
            for r in sheet.rules:
                if isinstance(r, Condition) and r.keyword == "import":
                    yield from get_filenames(r.style)
        return list(get_filenames(self))

    @property
    def style(self):
        """Return a Style object with the remaining rules.

        All rules that still are behind a condition, are let through.
        The rules are sorted on specificity.

        """
        def get_rules(rules):
            for r in rules:
                if isinstance(r, Condition):
                    yield from get_rules(r.style.rules)
                elif isinstance(r, Rule):
                    yield r
        rules = sorted(get_rules(self.rules),
            key=lambda rule: calculate_specificity(rule.prelude))
        rules.reverse()
        return Style(rules)

    @property
    def at(self):
        """Return an Atrules object containing the remaining at-rules.

        All rules that still are behind a condition, are let through.

        """
        def get_rules(rules):
            for r in rules:
                if isinstance(r, Condition):
                    yield from get_rules(r.style.rules)
                elif isinstance(r, Atrule):
                    yield r
        return Atrules(list(get_rules(self.rules)))


class Style:
    """Represents the list of rules created by the StyleSheet object.

    All ``select``-methods/properties return a new Style object with the
    narrowed-down selection of rules.

    Use ``properties()`` to get the dictionary of combined properties that
    apply to the selected rules.

    """
    def __init__(self, rules):
        self.rules = rules

    @style_query
    def select_element(self, element):
        """Select the rules that match with Element."""
        for rule in self.rules:
            if element.match(rule.prelude):
                yield rule

    @style_query
    def select_lxml_element(self, element):
        """Select the rules that match with lxml.etree.Element."""
        return self.select_element(LxmlElement(element))

    def properties(self):
        """Return the combined properties of the current set of rules. (Endpoint.)

        Returns a dictionary with the properties. The value of each property
        is a list of Value instances.

        """
        result = {}
        important_properties = set()
        for rule in self.rules:
            for key, value in rule.properties.items():
                important = False
                if value[-1] == '!important':
                    value = value[:-1]
                    important = True
                if key not in result:
                    result[key] = value
                elif important and key not in important_properties:
                    result[key] = value
                    important_properties.add(key)
        return result


class Atrules:
    """Represents the @rules that are not nested, e.g. @page etc."""
    def __init__(self, rules):
        self.rules = rules

    @style_query
    def select(self, *keywords):
        for r in self.rules:
            if r.keyword in keywords:
                yield r


class AbstractElement:
    """Base implementation for an Element object that Style uses for matching.

    You may reimplement this to wrap any tree structure you want to use with
    the css module. You should then implement:

     * ``__init__()``
     * ``get_name()``
     * ``get_parent()``
     * ``get_attributes()``
     * ``get_pseudo_classes()`` (if needed)
     * ``get_pseudo_elements()`` (if needed)
     * ``children()``
     * ``get_child_count()``
     * ``previous_siblings()``
     * ``next_siblings()``

    If you wrap other objects, be sure to reimplement __eq__ and __ne__, to
    compare those objects and not the wrappers, which may be recreated each
    time.

    """

    _pseudo_class = util.Dispatcher()

    def __bool__(self):
        """Always return True."""
        return True

    def __repr__(self):
        attrs = reprlib.repr(self.get_attributes())
        count = self.get_child_count()
        return "<Element {} {} ({} children)>".format(self.get_name(),
            attrs, count)

    def get_name(self):
        """Implement to return the element's name."""
        return ""

    def get_parent(self):
        """Implement to return the parent Element or None."""
        return None

    def get_attributes(self):
        """Implement to return a dictionary of attributes, keys and values are str."""
        return {}

    def get_pseudo_classes(self):
        """Implement to return a list of pseudo classes."""
        return []

    def get_pseudo_elements(self):
        """Implement to return a list of pseudo elements."""
        return []

    def children(self):
        """Implement to yield our children."""
        yield from ()

    def get_child_count():
        """Implement to return the number of children."""
        return 0

    def previous_siblings(self):
        """Implement to yield our previous siblings in backward order."""
        yield from ()

    def next_siblings(self):
        """Implement to yield our next siblings in forward order."""
        yield from ()

    def get_classes(self):
        """Return a tuple of classes, by default from the 'class' attribute.

        The returned tuple may be empty, when there are no class names.

        """
        d = self.get_attributes()
        if d:
            return d.get("class", "").split()
        return ()

    def get_id(self):
        """Return the id or None, by default read from the 'id' attribute."""
        d = self.get_attributes()
        if d:
            return d.get("id")

    def next_sibling(self):
        """Return the next sibling."""
        for e in self.next_siblings():
            return e

    def previous_sibling(self):
        """Return the previous sibling."""
        for e in self.previous_siblings():
            return e

    @_pseudo_class('first-child')
    def is_first_child(self):
        """Return True if we are the first child."""
        return not self.previous_sibling()

    @_pseudo_class('last-child')
    def is_last_child(self):
        """Return True if we are the last child."""
        return not self.next_sibling()

    @_pseudo_class('only-child')
    def is_only_child(self):
        """Return True if we are the only child."""
        return not self.next_sibling() and not self.previous_sibling()

    @_pseudo_class('first-of-type')
    def is_first_of_type(self):
        """Return True if we are the first of our type."""
        name = self.get_name()
        return not any(e.get_name() == name for e in self.previous_siblings())

    @_pseudo_class('last-of-type')
    def is_last_of_type(self):
        """Return True if we are the last of our type."""
        name = self.get_name()
        return not any(e.get_name() == name for e in self.next_siblings())

    @_pseudo_class('empty')
    def is_empty(self):
        """Return True if we have no child elements."""
        return not self.get_child_count()

    def match(self, prelude):
        """Match with a compound selector expression (``prelude`` part of Rule)."""
        for selectors in prelude:
            if not self.match_selectors(selectors):
                return False
        return True

    def match_selectors(self, selectors):
        """Match with a list of selectors with operators in between."""
        selectors = iter(reversed(selectors))
        sel = next(selectors, None)
        if not sel or not self.match_selector(sel):
            return False
        element = self
        operator = next(selectors, None)
        sel = next(selectors, None)
        while operator and sel:
            if operator == ">":
                # parent should match
                element = element.get_parent()
                if element is None or not element.match_selector(sel):
                    return False
            elif operator == " ":
                # an ancestor should match
                element = element.get_parent()
                while element is not None:
                    if element.match_selector(sel):
                        break
                    element = element.get_parent()
                else:
                    return False
            elif operator == "+":
                # immediate sibling should match
                for element in element.previous_siblings():
                    if element.match_selector(sel):
                        break
                    return False
                else:
                    return False
            else: # operator == "~":
                # a sibling should match
                for element in element.previous_siblings():
                    if element.match_selector(sel):
                        break
                else:
                    return False
            operator = next(selectors, None)
            sel = next(selectors, None)
        return True

    def match_selector(self, selector):
        """Match with a single CSS selector dictionary.

        Returns True if the element matches with the selector.

        """
        # class
        classes = self.get_classes()
        if any(c not in classes for c in selector.get('class_selector', ())):
            return False
        # element name?
        if any(n != self.get_name() for n in selector.get('element_selector', ())):
            return False
        # id?
        if any(i != self.get_id() for i in selector.get('id_selector', ())):
            return False
        # attrs?
        attributes = self.get_attributes()
        for attrname, operator, text, flag in selector.get('attribute_selector', ()):
            try:
                value = attributes[attrname]
            except KeyError:
                return False
            if operator and text:
                if flag in ('i', 'I'):
                    text = text.lower()
                    value = value.lower()
                if operator == "=":
                    if text != value:
                        return False
                elif operator == "~=":
                    if text not in value.split():
                        return False
                elif operator == "|=":
                    if text != value and not value.startswith(text + "-"):
                        return False
                elif operator == "^=":
                    if not value.startswith(text):
                        return False
                elif operator == "$=":
                    if not value.endswith(text):
                        return False
                elif operator == "*=":
                    if text not in value:
                        return False

        # pseudo_class?
        pseudo_classes = self.get_pseudo_classes()
        for c, selector_list in selector.get('pseudo_class', ()):
            result = self._pseudo_class(c)
            if result is None:
                if c not in pseudo_classes:
                    return False
            elif result is False:
                return False

        # pseudo_element?
        pseudo_elements = self.get_pseudo_elements()
        for c in selector.get('pseudo_element', ()):
            if c not in pseudo_elements:
                return False
        return True


class Element(AbstractElement, list):
    """Mimic an Element CSS selector rules are matched with.

    Use "class\_" when specifying the class with a keyword argument.
    You can also manipulate the attributes after instantiating.

    """
    def __init__(self, name="", parent=None, pseudo_classes=None, pseudo_elements=None, **attrs):
        super().__init__()
        self.name = name
        self.parent = parent
        self.pseudo_classes = pseudo_classes or []
        self.pseudo_elements = pseudo_elements or []
        self.attrs = attrs
        if "class" not in attrs and "class_" in attrs:
            attrs["class"] = attrs["class_"]
            del attrs["class_"]
        if parent:
            parent.append(self)

    def __eq__(self, other):
        return self is other

    def __ne__(self, other):
        return self is not other

    def get_name(self):
        """Implemented to return the element's name."""
        return self.name

    def get_parent(self):
        """Implemented to return the parent Element or None."""
        return self.parent

    def get_attributes(self):
        """Implemented to return a dictionary of attributes."""
        return self.attrs

    def get_pseudo_classes(self):
        """Implemented to return a list of pseudo classes."""
        return self.pseudo_classes

    def get_pseudo_elements(self):
        """Implemented to return a list of pseudo elements."""
        return self.pseudo_elements

    def children(self):
        """Implemented to yield our children."""
        yield from self

    def get_child_count(self):
        """Implemented to return the number of children."""
        return len(self)

    def previous_siblings(self):
        """Yield our previous siblings in backward order."""
        if self.parent is not None:
            i = self.parent.index(self)
            if i:
                yield from self.parent[i-1::-1]

    def next_siblings(self):
        """Yield our next siblings in forward order."""
        if self.parent is not None:
            i = self.parent.index(self)
            yield from self.parent[i+1:]


class LxmlElement(AbstractElement):
    """An Element wrapping an element from a lxml.etree tree."""
    def __init__(self, element):
        self.e = element

    def get_name(self):
        """Return the element's name."""
        return self.e.tag

    def get_parent(self):
        """Return the parent Element or None."""
        return type(self)(self.e.getparent())

    def get_attributes(self):
        """Return a dictionary of attributes, keys and values are str."""
        return self.e.attrib

    def get_pseudo_classes(self):
        """Implement to return a list of pseudo classes."""
        return []

    def get_pseudo_elements(self):
        """Implement to return a list of pseudo elements."""
        return []

    def children(self):
        """Yield our children."""
        for n in self.e:
            yield type(self)(n)

    def get_child_count(self):
        """Return the number of children."""
        return len(self.e)

    def previous_siblings(self):
        """Yield our previous siblings in backward order."""
        for n in self.e.itersiblings(preceding=True):
            yield type(self)(n)

    def next_siblings(self):
        """Yield our next siblings in forward order."""
        for n in self.e.itersiblings():
            yield type(self)(n)


def calculate_specificity(prelude):
    """Calculate the specificity of the Css rule prelude.

    Returns a three-tuple (ids, clss, elts), where ids is the number of ID
    selectors, clss the number of class, attribute or pseudo-class selectors,
    and elts the number of element or pseudo-elements.

    Currently, does not handle functions like :not(), :is(), although that
    would not be difficult to implement.

    """
    specificities = []
    total = lambda *names: sum(len(selector.get(n, ())) for n in names)
    for selectors in prelude:
        ids = clss = elts = 0
        for selector in selectors:
            if isinstance(selector, dict):
                ids += total('id_selector')
                clss += total('attribute_selector', 'class_selector', 'pseudo_class')
                elts += total('element_selector', 'pseudo_element')
        specificities.append((ids, clss, elts))
    return max(specificities)


def color2hex(color):
    """Return a hexadecimal string with '#' prepended for the Color instance."""
    r, g, b, a = color
    x = "#{:06x}".format(r*65536 + g*256 + b)
    if a < 1:
        x += format(int(a * 255), '02x')
    return x


def quote_if_needed(s):
    """Double-quote the string for CSS if it is not a valid ident-token."""
    if not re.fullmatch(r'[\w-]+', s):
        s = re.sub(r'[\\"]', lambda m: escape(m.group()), s)
        return '"' + s + '"'
    return s


def escape(char):
    """Escape the specified character for CSS."""
    return "".join(r"\{:x} ".format(ord(c)) for c in char)




# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
This module defines StandardAction.

A StandardAction is a singleton object. Acessing an attribute (without
underscore) creates that attribute as a new instance, with the current instance
as _parent.

This way a new action type can be created that shares its parent with other
types, a concept borrowed from pygments. An example::

    >>> Comment = StandardAction("Comment")
    >>> Literal = StandardAction("Literal")
    >>> String = Literal.String
    >>> String.DoubleQuotedString
    Literal.String.DoubleQuotedString
    >>> String.SingleQuotedString
    Literal.String.SingleQuotedString

StandardAction instances support iteration and membership methods.
Iteration yields the instance ifself and then the parents::

    >>> for i in String.DoubleQuoted:
    ...     print(i)
    ...
    Literal.String.DoubleQuoted
    Literal.String
    Literal

And the :keyword:`in` operator returns True when a standard action belongs to
another one, i.e. the other one is one of the ancestors of the current action::

    >>> String.DoubleQuotedString in String
    True
    >>> Literal in String
    False

The :keyword:`in` operator also works with a string::

    >>> 'String' in Literal.String.DoubleQuoted
    True
    >>> 'Literal' in String
    True

The last one could be surprising, but String is defined as ``Literal.String``::

    >>> String
    Literal.String

Finally, the `&` operator returns the common ancestor, if any::

    >>> String & Number
    Literal
    >>> String & Text
    >>>

See for the full list of pre-defined standard actions :doc:`stdactions`.


"""


import threading

# we use a global lock for standardaction creation, it seems overkill
# to me to equip every instance with one.
_lock = threading.Lock()

_toplevel_actions = {}       # store the "root" actions


class StandardAction:
    """Factory for standard action singletons."""
    def __new__(cls, name, parent=None):
        d = parent.__dict__ if parent else _toplevel_actions
        with _lock:
            try:
                return d[name]
            except KeyError:
                new = d[name] = object.__new__(cls)
                new._name = name
                new._parent = parent
                return new

    def __getattr__(self, name):
        if name.startswith('_'):
            raise AttributeError("{} has no attribute {}".format(self, repr(name)))
        return type(self)(name, self)

    def __repr__(self):
        return ".".join(reversed([n._name for n in self]))

    def __iter__(self):
        node = self
        while node:
            yield node
            node = node._parent

    def __contains__(self, other):
        if isinstance(other, str):
            return any(t._name == other for t in self)
        return any(t is self for t in other)

    def __and__(self, other):
        ancestors = frozenset(other)
        for t in self:
            if t in ancestors:
                return t

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self




# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Parse various troff dialects, most notably groff.
https://www.gnu.org/software/groff/

"""

__all__ = ('Troff',)

import re

from parce import Language, lexicon, skip, default_action, default_target
from parce.action import (
        Comment, Delimiter, Name, Number, Operator, String, Text)
from parce.rule import bygroup


class Troff(Language):
    @lexicon(re_flags=re.MULTILINE)
    def root(cls):
        yield r"^(['.])[ \t]*", bygroup(Delimiter.Request), cls.request, cls.name
        yield from cls.escapes()
        yield default_action, Text

    @lexicon(re_flags=re.MULTILINE)
    def request(cls):
        yield from cls.escapes()
        yield r'\n', skip, -1
        yield r' (")', bygroup(String), cls.string
        yield r'([-+]?(?:\d+(?:\.\d*)?)|\.\d+)([ciPpmMnuvsf])?', bygroup(Number, Name.Unit)
        yield r'[-+*/%=<>&:!()]|[=<>]=', Operator
        yield r'[<>]\?|;', Operator # GNU extension
        yield default_action, Text

    @lexicon
    def name(cls):
        """The name of a request, handle some special preprocessor macros."""
        yield r'lilypond\b', Name.Identifier.Preprocessed, -1, cls.preprocess_lilypond
        yield r'[^\W\d]\w*\b', Name.Identifier, -1
        yield default_target, -1

    @lexicon(re_flags=re.MULTILINE)
    def string(cls):
        """String in request arguments."""
        yield r'$', None, -1
        yield r'"', String, -1
        yield from cls.escapes(String)
        yield default_action, String

    @lexicon(re_flags=re.MULTILINE)
    def comment(cls):
        yield r'$', None, -1
        yield from cls.comment_common()

    @lexicon
    def preprocess_lilypond(cls):
        """If "start" is found, go to LilyPond, otherwise back to the request context."""
        yield r'[ \t]*(start)\b', bygroup(Name.Command), -2, cls.lilypond, cls.request
        yield default_target, -1 # back to request

    @lexicon(re_flags=re.MULTILINE)
    def lilypond(cls):
        """Stuff between .lilypond start and .lilypond end."""
        from .lilypond import LilyPond
        yield r"^(['.])[ \t]*(lilypond) +(end)\b", \
            bygroup(Delimiter.Request, Name.Identifier.Preprocessed, Name.Command), \
            -1
        yield from LilyPond.root

    @classmethod
    def escapes(cls, escape_base=Text):
        # escapes (man 7 groff)
        yield r'\\["#]', Comment, cls.comment
        # prevent interpreting request after continuation line
        yield r"(\\\n)([.'])?", bygroup(escape_base.Escape, Text)
        # single char escapes
        yield r"\\[\\'`_.%! 0|^&)/,~:{}acdeEprtu-]", escape_base.Escape
        # escapes expecting single, two-char or bracketed arg
        yield r"\\[\*$fFgkmMnVY](?:\[[^\]\n]*\]|\(..|.)", escape_base.Escape
        # \s
        yield r"\\s[-+]?(?:\d|\([-+]?\d\d|\[[\d+-]+\]|'[\d+-]+')", escape_base.Escape
        # escapes expecting a single-quoted argument
        yield r"\\[AbBCDlLnoRsSvwxXZ](?:'[^'\n]*')?", escape_base.Escape
        # \(<xx> or \[<xxxx>]
        yield r"\\(?:\[[^\]\n]*\]|\(..)", escape_base.Escape
        # undefined (last resort)
        yield r"\\", escape_base.Escape




# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Test the derived Lexicon(arg) stuff
"""

import sys
sys.path.insert(0, '.')

from parce import Language, lexicon, root
from parce.action import Name, Text
from parce.rule import arg, derive, MATCH


class MyLang(Language):
    @lexicon
    def root(cls):
        yield r"@([a-z]+)@", Name, derive(cls.here, MATCH(1))
        yield r"\w+", Text

    @lexicon
    def here(cls):
        yield arg(prefix=r"\b", suffix=r"\b"), Name, -1
        yield r"\w+", Text


def test_main():
    text = r""" text @mark@ bla bla mark bla bla """

    tree = root(MyLang.root, text)
    tree.dump()
    assert tree.query.all("mark").pick().is_last()
    assert tree.query.all(MyLang.here).pick()

    from parce.validate import validate_language
    assert validate_language(MyLang)


if __name__ == "__main__":
    test_main()




# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
This module provides the Language class, which serves as the base class
for all language definitions.

Additionally, there are some utility functions to manage and query built-in
language definitions.

"""

import importlib
import glob
import os

import parce
import parce.lang
from parce.action import Comment


class Language:
    """A Language represents a set of Lexicons comprising a specific language.

    A Language is never instantiated. The class itself serves as a namespace
    and can be inherited from.

    """

    @classmethod
    def comment_common(cls):
        """Provides subtle highlighting within comments.

        The default implementation highlights words like TODO, XXX, TEMP, etc.
        using Comment.Alert, and highlights URLs and email addresses with the
        Comment.Url and Comment.Email action respectively. Most bundled
        languages use this method for their comment lexicons.

        """
        yield r'\b\w+(?:[._%+-]\w+)*@\w+(?:[._-]\w+)*\b', Comment.Email
        yield r'(?:(?:https?|ftp):/|\bwww\.)(?:[\w_~:/#-]+([.?=][\w_~:/#-]+)*|\([\w._~:?/#-]*\))+', Comment.Url
        yield r"\b(ALERT|BUG|FIXME|TEMP|TODO|XXX+)\b", Comment.Alert
        yield parce.default_action, Comment


def get_all_modules():
    """Return the sorted list of module names in ``parce.lang``.

    Modules that start with an underscore are skipped.

    """
    names = []
    for filename in glob.glob(os.path.join(parce.lang.__path__[0], "*.py")):
        name = os.path.splitext(os.path.basename(filename))[0]
        if not name.startswith('_'):
            names.append(name)
    names.sort()
    return names


def get_languages(name):
    """Yield the Language subclasses defined in the module ``name``.

    The module name must be one of the modules returned by
    :meth:``get_all_modules``.

    """
    modname = 'parce.lang.' + name
    mod = importlib.import_module(modname)
    for name, obj in mod.__dict__.items():
        if (isinstance(obj, type) and issubclass(obj, parce.Language)
               and not obj.__name__.startswith('_')
               and obj is not parce.Language and obj.__module__ == modname):
            yield obj


def get_all_languages():
    """Import all modules in ``parce.lang`` and yield all defined classes
    that inherit from :class:`Language`.

    """
    for name in get_all_modules():
        yield from get_languages(name)





# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Tom's Obvious, Minimal Language.
https://github.com/toml-lang/toml
"""

__all__ = ('Toml',)

import re

from parce import Language, lexicon, skip, default_action, default_target
from parce.action import (
    Bracket, Comment, Delimiter, Invalid, Literal, Number, Name, Operator,
    Separator, String, Whitespace)
from parce.rule import TEXT, bygroup, call, select


# https://tools.ietf.org/html/rfc3339#section-5.6
RE_FULL_DATE = r"\d{4}-\d\d-\d\d"
RE_PARTIAL_TIME = r"\d\d:\d\d:\d\d(?:\.\d+)?"
RE_TIME_NUMOFFSET = r"[+-]\d\d:\d\d"
RE_TIME_OFFSET = r"(?:[zZ]|" + RE_TIME_NUMOFFSET + ")"
RE_FULL_TIME = RE_PARTIAL_TIME + RE_TIME_OFFSET + '?' # offset may be omitted
RE_DATE_TIME = RE_FULL_DATE + "[ tT]" + RE_FULL_TIME

RE_HEX = r'0[xX](?:_?[0-9a-fA-F])+'
RE_OCT = r'0[oO](?:_?[0-7])+'
RE_BIN = r'0[bB](?:_?[01])+'
RE_DEC = r'[-+]?\d(?:_?\d)*(?:\.(?:\d(?:_?\d)*)+)?(?:[eE][-+]?\d(?:_?\d)*)?'


class Toml(Language):
    @lexicon
    def root(cls):
        yield '#', Comment, cls.comment
        yield r'(\[\[)(?:[ \t]*(\.))?', bygroup(Bracket, Invalid), cls.array_table
        yield r'(\[)(?:[ \t]*(\.))?', bygroup(Bracket, Invalid), cls.table
        yield r'=[^\n#]*', Invalid
        yield r'\.[^\n#]*', Invalid
        yield r'\s+', skip
        yield default_target, cls.key

    @lexicon
    def table(cls):
        yield r'(?:(\.)[ \t]*)?(\])([^\n#]*)', \
            bygroup(Invalid, Bracket, select(call(str.isspace, TEXT), Invalid, skip)), -1
        yield from cls.keys()

    @lexicon
    def array_table(cls):
        yield r'(?:(\.)[ \t]*)?(\]\])([^\n#]*)', \
            bygroup(Invalid, Bracket, select(call(str.isspace, TEXT), Invalid, skip)), -1
        yield from cls.keys()

    @lexicon(re_flags=re.MULTILINE)
    def key(cls):
        yield '#', Comment, -1, cls.comment
        yield r'=', Operator.Assignment, -1, cls.value
        yield from cls.keys()

    @lexicon(re_flags=re.MULTILINE)
    def value(cls):
        yield '#', Comment, -1, cls.comment
        yield r'$', None, -1
        yield from cls.values()

    @classmethod
    def keys(cls):
        yield r'[A-Za-z0-9_-]+', Name.Variable
        yield r'''(\.)(?=[ \t]*[\}\],'"A-Za-z0-9_-])''', Delimiter.Dot
        yield r'"', String, cls.string_basic
        yield r"'", String, cls.string_literal
        yield r'[ \t]+', skip
        yield r'[^\s#=\]]+', Invalid

    @classmethod
    def values(cls):
        yield '#', Comment, cls.comment
        yield r'\[', Bracket, cls.array
        yield r'\{', Bracket, cls.inline_table
        yield r'"""', String, cls.string_multiline_basic
        yield r'"', String, cls.string_basic
        yield r"(''')(\n)?", bygroup(String, Whitespace), cls.string_multiline_literal
        yield r"'", String, cls.string_literal
        yield RE_DATE_TIME, Literal.Timestamp
        yield RE_FULL_DATE, Literal.Timestamp
        yield RE_FULL_TIME, Literal.Timestamp
        yield RE_OCT, Number
        yield RE_BIN, Number
        yield RE_HEX, Number
        yield RE_DEC, Number
        yield r"[-+]?\b(?:inf|nan)\b", Number
        yield r"\b(?:true|false)\b", Name.Constant
        yield r'\S+', Invalid

    @lexicon
    def array(cls):
        yield r'(\])([^,}#\n\]]*)', bygroup(Bracket, select(call(str.isspace, TEXT), Invalid, skip)), -1
        yield r',', Separator
        yield from cls.values()

    @lexicon
    def inline_table(cls):
        yield '#', Comment, cls.comment
        yield r'\}', Bracket, -1
        yield r'=', Operator.Assignment.Invalid
        yield r'\s+', skip
        yield default_target, cls.inline_key

    @lexicon
    def inline_key(cls):
        yield r'=', Operator.Assignment, -1, cls.inline_value
        yield r'\}', Bracket.Invalid, -1
        yield from cls.keys()

    @lexicon
    def inline_value(cls):
        yield '#', Comment, cls.comment
        yield r'\}', Bracket, -2
        yield r',', Separator, -1
        yield from cls.values()

    @lexicon
    def string_multiline_basic(cls):
        yield r'(""")([^\s,}#\]]*)', bygroup(String, Invalid), -1
        yield r'\\\s+', Whitespace
        yield r'\\(?:["\\bfnrt]|u[0-9a-fA-F]{4})', String.Escape
        yield r'\\.', String.Invalid
        yield default_action, String

    @lexicon(re_flags=re.MULTILINE)
    def string_basic(cls):
        yield r'(")([^\s,}#\]=]*)', bygroup(String, Invalid), -1
        yield r'\\(?:["\\bfnrt]|u[0-9a-fA-F]{4})', String.Escape
        yield r'\\.', String.Invalid
        yield r'[^"]*?$', String.Invalid, -1
        yield default_action, String

    @lexicon
    def string_multiline_literal(cls):
        yield r"(''')([^\s,}#\]]*)", bygroup(String, Invalid), -1
        yield default_action, String

    @lexicon(re_flags=re.MULTILINE)
    def string_literal(cls):
        yield r"(')([^\s,}#\]=]*)", bygroup(String, Invalid), -1
        yield r"[^']*?$", String.Invalid, -1
        yield default_action, String

    @lexicon(re_flags=re.MULTILINE)
    def comment(cls):
        yield from cls.comment_common()
        yield r'$', Comment, -1




# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


r"""
A Lexicon groups rules to match.

A Lexicon is created by decorating a method yielding rules with the
:attr:`@lexicon <lexicon>` decorator. (Although this actually
creates a LexiconDescriptor. When a LexiconDescriptor is accessed for the first
time via a Language subclass, a Lexicon for that class is created and cached,
and returned each time that attribute is accessed.)

The Lexicon can parse text according to the rules. When its :func:`parse`
function is called for the first time, the rules-function is run with the
language class as argument, and the rules it yields are cached.

This makes it possible to inherit from a Language class and only re-implement
some lexicons, the others keep working as in the base class.

Example:

    >>> from parce import Language, lexicon
    >>>
    >>> class MyLang(Language):
    ...     @lexicon
    ...     def numbers(cls):
    ...         yield r'\d+', "A number"
    ...         yield r'\w+', "A word"
    ...
    >>> MyLang.numbers
    MyLang.numbers
    >>> type(MyLang.numbers)
    <class 'parce.lexicon.Lexicon'>
    >>> for i in MyLang.numbers.parse("1 a2 d3 4 p 5", 0):
    ...  print(i)
    ...
    (0, '1', <re.Match object; span=(0, 1), match='1'>, 'A number', None)
    (2, 'a2', <re.Match object; span=(2, 4), match='a2'>, 'A word', None)
    (5, 'd3', <re.Match object; span=(5, 7), match='d3'>, 'A word', None)
    (8, '4', <re.Match object; span=(8, 9), match='4'>, 'A number', None)
    (10, 'p', <re.Match object; span=(10, 11), match='p'>, 'A word', None)
    (12, '5', <re.Match object; span=(12, 13), match='5'>, 'A number', None)

Parsing (better: lexing) is done by a :class:`~parce.lexer.Lexer` instance,
which switches Lexicon when a target is encountered.

"""

import itertools
import re
import threading

import parce.regex
from . import util
from .target import TargetFactory
from .ruleitem import (
    Item, RuleItem, evaluate_rule, needs_evaluation, pre_evaluate_rule)


class LexiconDescriptor:
    """The LexiconDescriptor creates a Lexicon when called via a class."""
    __slots__ = ('rules_func', '_lexicons', '_lock', '_re_flags')

    def __init__(self, rules_func,
                       re_flags=0,
        ):
        """Initializes with the rules function.

        The rules function accepts the Language class as argument, and yields
        the pattern, action, target, ... tuples.

        """
        self.rules_func = rules_func    #: the function yielding the rules
        self._re_flags = re_flags
        self._lexicons = {}
        self._lock = threading.Lock()

    def __get__(self, instance, owner):
        """Called when accessed as a descriptor, via the Language class."""
        if instance:
            raise RuntimeError('Language should never be instantiated')
        try:
            return self._lexicons[owner]
        except KeyError:
            # prevent instantiating the same Lexicon multiple times
            with self._lock:
                try:
                    lexicon = self._lexicons[owner]
                except KeyError:
                    lexicon = self._lexicons[owner] = Lexicon(self, owner)
                return lexicon


class Lexicon:
    """A Lexicon parses text according to rules.

    A Lexicon is tied to a particular class, which makes it possible to inherit
    from a Language class and change only some Lexicons.

    .. py:function:: parse(text, pos)

        Start parsing ``text`` from the specified position.
        Yields five-tuples ``(pos, text, matchobj, action, target)``.

        The ``pos`` is the start position a match was found, ``text`` is the
        matched text, ``matchobj`` the match object (which can be None for
        default actions), ``action`` the action that was specified in the
        matching rule, and ``target`` is either None or a
        :class:`~parce.target.Target` object.

    """
    def __init__(self, descriptor, language, arg=None):
        #: The LexiconDescriptor this Lexicon was created by.
        self.descriptor = descriptor
        #: The Language class the lexicon belongs to.
        self.language = language
        #: The argument the lexicon was called with (creating a derived
        #: Lexicon). None for a normal lexicon.
        self.arg = arg
        #: The short name (name of the method this Lexicon was defined with)
        self.name = descriptor.rules_func.__name__
        #: The short name with the Language name prepended, like
        #: ``'Language.lexicon'``.
        self.fullname = language.__name__ + '.' + self.name
        #: The full name with the Language's module prepended, like
        #: ``'parce.lang.xml.Xml.root'``.
        self.qualname = language.__module__ + '.' + self.fullname
        self.__doc__ = descriptor.rules_func.__doc__
        self._derived = {}
        # lock is used when creating a derivate and/or the parse() instance function
        self._lock = threading.Lock()

    def __call__(self, arg=None):
        """Create a derived Lexicon with argument ``arg``.

        The argument should be a simple, hashable singleton object, such as a
        string, an integer or a standard action. The created Lexicon is cached.
        The argument is accessible using special pattern and rule item types,
        so a derived Lexicon can parse text based on rules that are defined at
        parse time, which is useful for things like here documents, where you
        only get to know the end token after the start token has been found.

        When comparing Lexicons with ``equals()``, a derived lexicon compares
        equal with the Lexicon that created them, although they co-exist as
        separate objects.

        When yielding the rules from a derived lexicon, the dynamic rule items
        that depend on the Lexicon argument are already evaluated. When
        yielding the rules from a vanilla lexicon, they are not evaluated, so
        they adjust themselves to the lexicon they are included in (which will
        then evaluate the rules of course).

        If arg is None, self is returned.

        """
        if arg is None:
            return self
        elif self.arg is not None:
            vanilla = self.descriptor.__get__(None, self.language)
            return vanilla(arg)
        try:
            return self._derived[arg]
        except KeyError:
            with self._lock:
                try:
                    lexicon = self._derived[arg]
                except KeyError:
                    lexicon = self._derived[arg] = Lexicon(self.descriptor, self.language, arg)
            return lexicon

    def equals(self, other):
        """Return True if we are the same lexicon or a derivate from the same."""
        return self.descriptor is other.descriptor and self.language is other.language

    @util.cached_property
    def _rules(self):
        """Yield the rules.

        Rule items that depend on the lexicon argument are not yet evaluated.

        """
        return tuple(self.descriptor.rules_func(self.language) or ())

    @util.cached_property
    def rules(self):
        """Return all rules in a tuple.

        Rule items that depend on the lexicon argument are already evaluated.

        """
        return tuple(pre_evaluate_rule(rule, self.arg)
                for rule in self.descriptor.rules_func(self.language) or ())

    def __iter__(self):
        """Yield the rules.

        Patterns are created when this method is called for the first time. If
        this is a derived lexicon, dynamic rule items that depend on the
        argument are already evaluated.

        """
        yield from self.rules if self.arg else self._rules

    def __repr__(self):
        s = self.fullname
        if self.arg:
            s += '*'
        return s

    def __getattr__(self, name):
        """Create certain instance attributes when requested the first time.

        Calls :meth:`get_instance_attributes` to get instance attributes needed
        to use the Lexicon. Those attributes then are set in the Lexicon
        instance, so the do not need to be computed again.

        """
        if name in ("parse",):
            with self._lock:
                try:
                    return object.__getattribute__(self, name)
                except AttributeError:
                    self.parse = self._get_instance_attributes()
        return object.__getattribute__(self, name)

    @property
    def re_flags(self):
        """The re_flags that were set on instantiation."""
        return self.descriptor._re_flags

    def _get_instance_attributes(self):
        """Compile the pattern rules and return instance attributes.

        These are:

        ``parse``
            A ``parse(text, pos)`` function that parses text.

        """
        patterns = []
        rules = []
        no_default_action = object()
        default_action = no_default_action
        default_target = None

        make_target = TargetFactory.make

        # make lists of pattern, action and possible targets
        for pattern, *rule in self.rules:
            if pattern is parce.default_action:
                default_action = rule[0]
            elif pattern is parce.default_target:
                default_target = make_target(self, rule)
            elif rule and pattern is not None and pattern not in patterns:
                # skip rule when the pattern is None or already seen
                patterns.append(pattern)
                rules.append(rule)

        # prepare to handle a dynamic default action
        if isinstance(default_action, RuleItem):
            def dynamic_default_action(text):
                return default_action.evaluate({'text': text})
        else:
            dynamic_default_action = False

        # handle the empty lexicon case
        if not patterns:
            if dynamic_default_action:
                def parse(text, pos):
                    t = text[pos:]
                    yield pos, t, None, dynamic_default_action(t), None
            elif default_action is not no_default_action:
                def parse(text, pos):
                    yield pos, text[pos:], None, default_action, None
            elif default_target:
                def parse(text, pos):
                    if pos < len(text):
                        yield pos, "", None, None, default_target
            else:
                # just quits parsing
                def parse(text, pos):
                    yield from ()
            return parse

        # if there is only one pattern, and no dynamic action or target,
        # see if the pattern is simple enough to just use str.find
        if len(patterns) == 1 and not self.re_flags & re.IGNORECASE and \
                not needs_evaluation(rules[0]):
            needle = parce.regex.to_string(patterns[0])
            if needle:
                l = len(needle)
                action, *rule = rules[0]
                target = make_target(self, rule)
                if dynamic_default_action:
                    def parse(text, pos):
                        """Parse text, using a default action for unknown text."""
                        while True:
                            i = text.find(needle, pos)
                            if i > pos:
                                t = text[pos:i]
                                yield pos, t, None, dynamic_default_action(t), None
                            elif i == -1:
                                break
                            yield i, needle, None, action, target
                            pos = i + l
                        if pos < len(text):
                            t = text[pos:]
                            yield pos, t, None, dynamic_default_action(t), None
                elif default_action is not no_default_action:
                    def parse(text, pos):
                        """Parse text, using a default action for unknown text."""
                        while True:
                            i = text.find(needle, pos)
                            if i > pos:
                                yield pos, text[pos:i], None, default_action, None
                            elif i == -1:
                                break
                            yield i, needle, None, action, target
                            pos = i + l
                        if pos < len(text):
                            yield pos, text[pos:], None, default_action, None
                elif default_target:
                    def parse(text, pos):
                        """Parse text, stopping with the default target at unknown text."""
                        while needle == text[pos:pos+l]:
                            yield pos, needle, None, action, target
                            pos += l
                        if pos < len(text):
                            yield pos, "", None, None, default_target
                else:
                    def parse(text, pos):
                        """Parse text, skipping unknown text."""
                        while True:
                            i = text.find(needle, pos)
                            if i == -1:
                                break
                            yield i, needle, None, action, target
                            pos = i + l
                return parse

        # compile the regexp for all patterns
        rx = re.compile("|".join("(?P<g_{0}>{1})".format(i, pattern)
            for i, pattern in enumerate(patterns)), self.re_flags)
        # make a fast mapping list from matchObj.lastindex to the rules.
        # rules that contain Item instances are put in the dynamic index
        indices = sorted(v for k, v in rx.groupindex.items() if k.startswith('g_'))
        static = [None] * (indices[-1] + 1)
        dynamic = [None] * (indices[-1] + 1)
        for i, rule in zip(indices, rules):
            if needs_evaluation(rule):
                dynamic[i] = rule
            else:
                action, *target = rule
                static[i] = (action, make_target(self, target))

        # for rule containing no dynamic stuff, static has the rule, otherwise
        # falls back to dynamic, which is then immediately executed
        def token(m):
            """Return pos, text, match, *rule for the match object."""
            return (m.start(), m.group(), m, *(static[m.lastindex] or replace(m)))

        def replace(m):
            """Recursively replace dynamic rule items in the rule pointed to by match object."""
            action, *target = evaluate_rule(dynamic[m.lastindex], m)
            return action, make_target(self, target)

        if dynamic_default_action:
            finditer = rx.finditer
            def parse(text, pos):
                """Parse text, using a default action for unknown text."""
                for m in finditer(text, pos):
                    if m.start() > pos:
                        t = text[pos:m.start()]
                        yield pos, t, None, dynamic_default_action(t), None
                    yield token(m)
                    pos = m.end()
                if pos < len(text):
                    t = text[pos:]
                    yield pos, t, None, dynamic_default_action(t), None
        elif default_action is not no_default_action:
            finditer = rx.finditer
            def parse(text, pos):
                """Parse text, using a default action for unknown text."""
                for m in finditer(text, pos):
                    if m.start() > pos:
                        yield pos, text[pos:m.start()], None, default_action, None
                    yield token(m)
                    pos = m.end()
                if pos < len(text):
                    yield pos, text[pos:], None, default_action, None
        elif default_target:
            match = rx.match
            def parse(text, pos):
                """Parse text, stopping with the default target at unknown text."""
                while True:
                    m = match(text, pos)
                    if m:
                        yield token(m)
                        pos = m.end()
                    else:
                        if pos < len(text):
                            yield pos, "", None, None, default_target
                        break
        else:
            finditer = rx.finditer
            def parse(text, pos):
                """Parse text, skipping unknown text."""
                return map(token, finditer(text, pos))
        return parse


def lexicon(rules_func=None, **kwargs):
    """Lexicon factory decorator.

    Use this decorator to make a function in a Language class definition a
    LexiconDescriptor object. The LexiconDescriptor is a descriptor, and when
    calling it via the Language class attribute, a Lexicon is created, cached
    and returned.

    You can specify keyword arguments, that will be passed on to the Lexicon
    object as soon as it is created.

    The following keyword arguments are supported:

    ``re_flags``:
        The flags that are passed to the regular expression compiler

    The code body of the function should return (yield) the rules of the
    lexicon, and is run with the Language class as first argument, as soon as
    the lexicon is used for the first time.

    You can also call the Lexicon object just as an ordinary classmethod, to
    get the rules, e.g. for inclusion in a different lexicon.

    """
    if rules_func and not kwargs:
        return LexiconDescriptor(rules_func)
    def lexicon(rules_func):
        return LexiconDescriptor(rules_func, **kwargs)
    return lexicon

